#include <chrono>
#include <cmath>
#include <cstdint>
#include <mavsdk/mavsdk.h>
#include <mavsdk/plugins/action/action.h>
#include <mavsdk/plugins/manual_control/manual_control.h>
#include <mavsdk/plugins/mavlink_passthrough/mavlink_passthrough.h>
#include <mavsdk/plugins/telemetry/telemetry.h>
#include <iostream>
#include <fstream>
#include <functional>
#include <future>
#include <memory>
#include <thread>

#include "mavsdk_wrapper.h"

using namespace mavsdk;
using std::chrono::seconds;
using std::this_thread::sleep_for;

#define ERROR_CONSOLE_TEXT "\033[31m" // Turn text on console red
#define NORMAL_CONSOLE_TEXT "\033[0m" // Restore normal console colour

static std::ofstream log_file_fd;

static Mavsdk _mavsdk;
static Telemetry * telemetry;
static Action * action;
static ManualControl * manual_control;
static MavlinkPassthrough * mavlink_passthrough;
static std::shared_ptr<System> msystem;

static auto prom = std::promise<std::shared_ptr<System>>{};
static std::future<std::shared_ptr<System>> fut;

static Telemetry::FlightMode flight_mode;
static Telemetry::EulerAngle angle;
static Telemetry::FixedwingMetrics metric;
static Telemetry::Position position;
static uint64_t timestamp_us;

static const double EARTH_RADIUS = 6371000.F;

static bool autocontinue = false;
static double previousBearing;
static double xy_ratio;
static double target_latitude;
static double target_longitude;
static int mavsdk_started = 0;
static Telemetry::Position origin;

// Logs functions

static void log(std::string message) {
    log_file_fd << message << std::endl;
}

static void log_error(std::string message) {
    log(ERROR_CONSOLE_TEXT + message + NORMAL_CONSOLE_TEXT);
}

template <class Enumeration>
static void log_error_from_result(std::string message, Enumeration result) {
  std::ostringstream oss;
  oss << message << ": " << result;
  log_error(oss.str());
}

// Action functions

static int doAction(Action::Result (Action::*toDo)(void) const, std::string failure_message) {
    if(!mavsdk_started) {
        log_error("Mavsdk not started");
        return -1;
    }

    const Action::Result result = (action->*toDo)();
    if(result != Action::Result::Success) {
        log_error_from_result(failure_message, result);
        return -1;
    }
    return 0;
}

static int doMavlinkCommand(MavlinkPassthrough::CommandLong command, std::string failure_message) {
    const MavlinkPassthrough::Result result = mavlink_passthrough->send_command_long(command);
    if(result != MavlinkPassthrough::Result::Success) {
        log_error_from_result(failure_message, result);
        return -1;
    }
    return 0;
}

// Connexion management functions

static double toRad(double angle) {
    return M_PI * angle / 180;
}

int mavsdk_start(const char * url, const char * log_file, int timeout) {
    std::string connection_url(url);
    ConnectionResult connection_result;
    log_file_fd.open(log_file);

    connection_result = _mavsdk.add_any_connection(connection_url);
    if (connection_result != ConnectionResult::Success) {
        log_error_from_result("Connection failed", connection_result);
        return -1;
    }

    log("Waiting to discover msystem...");
    fut = prom.get_future();

    _mavsdk.subscribe_on_new_system([]() {
        auto msystem_tmp = _mavsdk.systems().back();

        if (msystem_tmp->has_autopilot()) {
            log("Discovered autopilot");

            // Unsubscribe again as we only want to find one system.
            _mavsdk.subscribe_on_new_system(nullptr);
            prom.set_value(msystem_tmp);
        }
    });

    if (fut.wait_for(seconds(timeout)) == std::future_status::timeout) {
        log_error("No autopilot found, exiting.");
        return -1;
    }

    msystem = fut.get();
    telemetry = new Telemetry(msystem);
    action = new Action(msystem);
    manual_control = new ManualControl(msystem);
    mavlink_passthrough = new MavlinkPassthrough(msystem);

    log("Subscribing to flight mode...");
    // Subscribe to receive updates on flight mode. You can find out whether FollowMe is active.
    telemetry->subscribe_flight_mode([](Telemetry::FlightMode _flight_mode) {
        if(flight_mode != _flight_mode) {
            flight_mode = _flight_mode;
        }
    });

    log("Subscribing to Euler angle...");
    telemetry->subscribe_attitude_euler([](Telemetry::EulerAngle euler_angle) {
        angle = euler_angle;
    });

    log("Subscribing to Fixedwing Metrics...");
    telemetry->subscribe_fixedwing_metrics([](Telemetry::FixedwingMetrics m) {
        metric = m;
    });

    log("Subscribing to position...");
    telemetry->subscribe_position([](Telemetry::Position pos) {
        timestamp_us = std::chrono::duration_cast<std::chrono::milliseconds>(
                          std::chrono::system_clock::now().time_since_epoch()
                       ).count();

        std::ostringstream oss;
	      position = pos;

        oss << mavsdk_getTimestamp() << "; "
            << mavsdk_getLatitude() << "; " << mavsdk_getLongitude() << "; "
            << mavsdk_getAltitude() << "; " << mavsdk_getAltitudeRel() << "; "
            << mavsdk_getPitch() << "; " << mavsdk_getRoll() << "; " << mavsdk_getYaw() << "; "
            << mavsdk_getAirspeed() << "; " << mavsdk_getThrottle() << "; " << mavsdk_getClimbRate();
        log(oss.str());
    });

    while(isnan(position.absolute_altitude_m)) {
        log("Waiting for first telemetry");
        sleep_for(seconds(1));
    }
    origin = position;
    xy_ratio = std::cos(toRad(origin.latitude_deg));

    log("MAVSDK started...");
    mavsdk_started = 1;
    log("timestamp; latitude; longitude; AMSL (m); rel altitude (m); pitch (°); roll(°); yaw(°); air speed (m/s); throttle(%); climb rate(m/s)");
    return 0;
}

int mavsdk_stop() {
    if(!mavsdk_started)
        return -1;
    
    if(!mavsdk_landed()) {
      log_error("You must land first !");
      return -1;
    }

    if(doAction(&Action::shutdown, "Shutdown failed")) {
        return -1;
    }

    // Delete pointers
    delete action;
    delete manual_control;
    delete mavlink_passthrough;
    delete telemetry;
    log_file_fd.close();

    return 0;
}

int mavsdk_reboot() {
    return doAction(&Action::reboot, "Rebooting failed");
}

// Flight state management functions

int mavsdk_arm(void) {
    if(!mavsdk_started)
        return -1;

    while(!telemetry->health().is_home_position_ok) {
        log("Waiting for home position to be set");
        sleep_for(seconds(1));
    }

    log("Arming...");
    return doAction(&Action::arm, "Arming failed");
}

int mavsdk_land(void) {
    log("Landing...");
    if(doAction(&Action::terminate, "Land failed")) {
        return -1;
    }

    // Check if vehicle is still in air
    while(telemetry->in_air()) {
        log("Vehicle is landing...");
        sleep_for(seconds(1));
    }
    log("Landed!");

    while(telemetry->armed()) {
        sleep_for(seconds(1));
    }
    log("Finished...");

    return 0;
}

int mavsdk_takeOff(void) {
    if(doAction(&Action::takeoff, "Takeoff failed")) {
        return -1;
    }

    while(flight_mode != Telemetry::FlightMode::Takeoff) {
        sleep_for(seconds(1));
    }

    log("Taking off...");
    return 0;
}

int mavsdk_takeOffAndWait(void) {
    if(mavsdk_takeOff() < 0) {
        return -1;
    }

    while(flight_mode == Telemetry::FlightMode::Takeoff) {
        sleep_for(seconds(1));
    }
    return 0;
}

int mavsdk_triggerParachute(void) {
    if (!mavsdk_started)
        return -1;

    MavlinkPassthrough::CommandLong command;
    command.command = MAV_CMD_DO_PARACHUTE;
    command.param1 = 2; //see https://mavlink.io/en/messages/common.html#PARACHUTE_ACTION
    command.target_sysid = mavlink_passthrough->get_target_sysid();
    command.target_compid = mavlink_passthrough->get_target_compid();
    return doMavlinkCommand(command, "Parachute release failed");
}

// Flight management functions

int mavsdk_loiter(float radius) {
    if (!mavsdk_started)
        return -1;
    Telemetry::Position position = telemetry->position();

    MavlinkPassthrough::CommandLong command;
    command.command = MAV_CMD_DO_REPOSITION;
    command.param1 = -1; // Ground speed, -1 for default
    command.param2 = MAV_DO_REPOSITION_FLAGS_CHANGE_MODE; //will go to navigate mode
    command.param3 = radius; // loiter radius
    command.param5 = (float)position.latitude_deg;
    command.param6 = (float)position.longitude_deg;
    command.param7 = position.absolute_altitude_m; //altitude is ignored, use altitude override package
    command.target_sysid = mavlink_passthrough->get_target_sysid();
    command.target_compid = mavlink_passthrough->get_target_compid();

    return doMavlinkCommand(command, "Entering loiter mode failed");
}

int mavsdk_setAirspeed(float airspeed) {
    if(!mavsdk_started)
        return -1;

    MavlinkPassthrough::CommandLong command;
    command.command = MAV_CMD_DO_OVERRIDE;
    command.param1 = 1 | 2 | 4 | 8;
    command.param2 = 2 | 4 | 8;
    command.param4 = airspeed;
    command.target_sysid = mavlink_passthrough->get_target_sysid();
    command.target_compid = mavlink_passthrough->get_target_compid();

    return doMavlinkCommand(command, "Setting airspeed failed");
}

int mavsdk_setAltitude(float altitude) {
    if (!mavsdk_started)
        return -1;

    MavlinkPassthrough::CommandLong command;
    command.command = MAV_CMD_DO_OVERRIDE;
    command.param1 = 1 | 2 | 4 | 8;
    command.param2 = 1 | 2 | 8;
    command.param3 = altitude;
    command.target_sysid = mavlink_passthrough->get_target_sysid();
    command.target_compid = mavlink_passthrough->get_target_compid();

    return doMavlinkCommand(command, "Setting altitude failed");
}

static int startAltitudeControl(void) {
    const ManualControl::Result result = manual_control->start_altitude_control();
    if (result != ManualControl::Result::Success) {
        log_error_from_result("Set manual control failed!", result);
        return -1;
    }
    return 0;
}

static double bearing(double lat1, double lon1, double lat2, double lon2) {
    double lat1_rad = toRad(lat1);
    double lat2_rad = toRad(lat2);

    double dL = toRad(lon2 - lon1);
    double x = cos(lat2_rad) * sin(dL);
    double y = cos(lat1_rad) * sin(lat2_rad) - sin(lat1_rad) * cos(lat2_rad) * cos(dL);
    return atan2(x, y) * 180 / M_PI;
}

int mavsdk_setTargetCoordinates(double la, double lo, float a) {
    int result = 0;
    Telemetry::Position position = telemetry->position();

    if (!mavsdk_started)
        return -1;

    target_latitude = la;
    target_longitude = lo;
    previousBearing = bearing(position.latitude_deg, position.longitude_deg,
                              la, lo);
    autocontinue = false;

    if (!mavsdk_isInManualMode()) {
        result = mavsdk_setManualControlInput();
        result |= startAltitudeControl();
    }
    result |= mavsdk_setAltitude(a);
    return result;
}

static int setManualControlInput(float x, float y, float z, float r) {
    const ManualControl::Result result = manual_control->set_manual_control_input(x, y, z, r);
    if(result != ManualControl::Result::Success) {
        log_error_from_result("Set manual control input failed!", result);
        return -1;
    }
    return 0;
}

int mavsdk_setManualControlInput(void) {
    if (autocontinue)
        return setManualControlInput(0, 0, 0, 0);

    Telemetry::Position position = telemetry->position();
    double b = bearing(position.latitude_deg, position.longitude_deg,
                       target_latitude, target_longitude);
    if (abs(b - previousBearing) > 160) {
        autocontinue = true;
    } else {
        previousBearing = b;
    }

    float angle_diff = (float)b - mavsdk_getYaw();
    return setManualControlInput(0, (angle_diff > 0 ? 1 : -1) * std::min(abs(angle_diff), 30.0f) / 30, 0, 0);
}

// Information functions

float mavsdk_getAltitude(void) {
    return position.absolute_altitude_m;
}

float mavsdk_getAltitudeRel(void) {
    return position.relative_altitude_m;
}

float mavsdk_getInitialAltitude(void) {
    return origin.absolute_altitude_m;
}

double mavsdk_getInitialLatitude(void) {
    return origin.latitude_deg;
}

double mavsdk_getInitialLongitude(void) {
    return origin.longitude_deg;
}

double mavsdk_getLatitude(void) {
    return position.latitude_deg;
}

double mavsdk_getLongitude(void) {
    return position.longitude_deg;
}

uint64_t mavsdk_getTimestamp(void) {
    return timestamp_us;
}

double mavsdk_getTakeOffAltitude(void) {
    const std::pair<Action::Result, float> response = action->get_takeoff_altitude();

    if(response.first != Action::Result::Success) {
        log_error_from_result("Get takeoff altitude failed", response.first);
        return -1;
    }
    return response.second;
}

float mavsdk_getPitch(void) {
    return angle.pitch_deg;
}

float mavsdk_getRoll(void) {
    return angle.roll_deg;
}

float mavsdk_getYaw(void) {
    return angle.yaw_deg;
}

float mavsdk_getAirspeed(void) {
    return metric.airspeed_m_s;
}

float mavsdk_getClimbRate(void) {
    return metric.climb_rate_m_s;
}

float mavsdk_getThrottle(void) {
    return metric.throttle_percentage;
}

int mavsdk_healthAllOk(void) {
    return telemetry->health_all_ok();
}

bool mavsdk_isInManualMode(void) {
    return telemetry->flight_mode() == Telemetry::FlightMode::Altctl;
}

int mavsdk_landed(void) {
    return !telemetry->in_air();
}
